{
  description = "AVR";

  inputs = {
    nixpkgs.url = "nixpkgs/nixos-21.11";
  };

  outputs = { self, nixpkgs }:
    let 
      version = builtins.string 0 8 self.lastModifiedDate;
      supportedSystems = [ "x86_64-linux" "avr-none" ];
      mcu_cflags = { mcu, chip_select, f_cpu }: ''-O -finline-functions -funsigned-char -funsigned-bitfields -fpack-struct -fshort-enums -Wall -Werror -Wstrict-prototypes -DF_CPU=${f_cpu} -std=gnu99 -D__${chip_select}__=1 -mmcu=${mcu}'';

      attiny85 = { mcu = "attiny85"; chip_select = "AVR_ATtiny85"; f_cpu = "8000000"; };
      atmega32u4 = { mcu = "atmega32u4"; chip_select = "AVR_ATmega32u4"; f_cpu = "8000000"; };
      processor = atmega32u4;

      mkLibrary = { pkgs, gcc, cflags, pname, psrc, pbuildInputs ? [] }:
      pkgs.stdenv.mkDerivation rec {
        name = pname;
        src = psrc;

        buildInputs = pbuildInputs;

        include_dirs = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}/include") buildInputs);
        object_files = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);

        dontUnpack = true;
        dontConfigure = true;

        buildPhase = ''
          for source in ${src}/*.c; do
              if [ -e $source ]; then
                  ${gcc} ${cflags} ${include_dirs} -c $source
              fi
          done
          cp ${src}/*.h . || true
          '';

        installPhase = ''
          mkdir -p $out/include $out/lib
          cp *.h $out/include || true
          cp *.o $out/lib || true
          '';
      };

      mkProgram = { pkgs, gcc, cflags, pname, psrc, pbuildInputs ? [] }:
      pkgs.stdenv.mkDerivation rec {
        name = pname;
        src = psrc;

        buildInputs = pbuildInputs;

        include_dirs = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}/include") buildInputs);
        object_files = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/lib/*.o") buildInputs);

        dontUnpack = true;
        dontConfigure = true;

        buildPhase = ''
          set -x
          for source in ${src}/*.c; do
              if [ -e $source ]; then
                  ${gcc} ${cflags} ${include_dirs} -c $source
              fi
          done

          ${gcc} -o ${name} ${cflags} ${object_files} *.o
          '';

        installPhase = ''
          mkdir -p $out/bin
          cp ${name} $out/bin
          '';
      };


    in rec
    {
      defaultPackage."x86_64-linux" = 
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.symlinkJoin {
          name = "avr-monorepo";
          paths = [
            (packages."x86_64-linux"."prime-tx" { gcc = "${avr.gcc}/bin/avr-gcc"; cflags = mcu_cflags atmega32u4; })
            (packages."x86_64-linux"."radio-rx" { gcc = "${avr.gcc}/bin/avr-gcc"; cflags = mcu_cflags atmega32u4; })
          ];
        };

      /* packages."x86_64-linux"."dio-attiny85" = (packages."x86_64-linux"."dio" attiny85); */

      /*
      packages."x86_64-linux"."dio" =
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          mkDerivation = pkgs.stdenv.mkDerivation;
          gcc =
          cflags = 
          name =
      */

      packages."x86_64-linux"."dio" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "dio";
          psrc = ./dio;
        };

      packages."x86_64-linux"."rng" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "rng";
          psrc = ./rng;
        };

      packages."x86_64-linux"."animation" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "animation";
          psrc = ./animation;
        };

      packages."x86_64-linux"."sk9822" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "sk9822";
          psrc = ./sk9822;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."shift-register" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "shift-register";
          psrc = ./shift-register;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."spi" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "spi";
          psrc = ./spi;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."rfm" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "rfm";
          psrc = ./rfm69hcw;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
            (packages."x86_64-linux"."spi" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."display" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkLibrary {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "display";
          psrc = ./display;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
            (packages."x86_64-linux"."shift-register" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."prime-tx" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkProgram {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "prime-tx";
          psrc = ./prime-tx;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
            (packages."x86_64-linux"."spi" { inherit gcc cflags; })
            (packages."x86_64-linux"."shift-register" { inherit gcc cflags; })
            (packages."x86_64-linux"."rfm" { inherit gcc cflags; })
            (packages."x86_64-linux"."display" { inherit gcc cflags; })
          ];
        };

      packages."x86_64-linux"."radio-rx" =
        { gcc, cflags }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
        in mkProgram {
          pkgs = pkgs;
          gcc = gcc;
          cflags = cflags;
          pname = "radio-rx";
          psrc = ./radio-rx;
          pbuildInputs = [
            (packages."x86_64-linux"."dio" { inherit gcc cflags; })
            (packages."x86_64-linux"."spi" { inherit gcc cflags; })
            (packages."x86_64-linux"."shift-register" { inherit gcc cflags; })
            (packages."x86_64-linux"."rfm" { inherit gcc cflags; })
            (packages."x86_64-linux"."display" { inherit gcc cflags; })
          ];
        };

        /*
      packages."x86_64-linux"."animation-test" =
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
        in pkgs.stdenv.mkDerivation rec {
          name = "animation-test";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."animation" { inherit gcc; })
          ];

          INCLUDE_DIRS = lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);
          gcc = "${pkgs.gcc}/bin/gcc";

          buildPhase = ''
            ${gcc} ${INCLUDE_DIRS} -o main.o -c ${src}/animation/main.c
            ${gcc} ${INCLUDE_DIRS} -o animation-test main.o ${OBJECT_FILES}
            '';
          installPhase = ''
            mkdir $out
            cp animation-test $out
            '';
        };

      packages."x86_64-linux"."flame" =
        { mcu, chip_select, f_cpu }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.stdenv.mkDerivation rec {
          name = "flame";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."dio" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."sk9822" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."rng" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."animation" { inherit mcu chip_select f_cpu;
                                                   gcc = "${avr.gcc}/bin/avr-gcc"; })
          ];

          CFLAGS = mcu_cflags { inherit mcu chip_select f_cpu; };
          INCLUDE_DIRS = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);

          buildPhase = ''
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o main.o -c ${src}/flame/main.c
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o flame.elf main.o ${OBJECT_FILES}
            $OBJCOPY -O ihex flame.elf flame.hex
            '';
          installPhase = ''
            mkdir $out
            cp flame.elf flame.hex $out
            '';
        };

      packages."x86_64-linux"."prime-tx" =
        { mcu, chip_select, f_cpu }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.stdenv.mkDerivation rec {
          name = "prime-tx";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."dio" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."spi" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."shift-register" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."rfm" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."display" { inherit mcu chip_select f_cpu; })
          ];

          CFLAGS = mcu_cflags { inherit mcu chip_select f_cpu; };
          INCLUDE_DIRS = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);


          buildPhase = ''
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o main.o -c ${src}/prime-tx/main.c
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} -o prime-tx.elf main.o ${OBJECT_FILES}
            $OBJCOPY -O ihex prime-tx.elf prime-tx.hex
            '';

          installPhase = ''
            mkdir $out
            cp prime-tx.elf prime-tx.hex $out
            '';
        };

      packages."x86_64-linux"."radio-rx" =
        { mcu, chip_select, f_cpu }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.stdenv.mkDerivation rec {
          name = "radio-rx";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."dio" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."spi" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."shift-register" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."rfm" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."display" { inherit mcu chip_select f_cpu; })
          ];

          CFLAGS = mcu_cflags { inherit mcu chip_select f_cpu; };
          INCLUDE_DIRS = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);

          buildPhase = ''
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o main.o -c ${src}/radio-rx/main.c
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} -o radio-rx.elf main.o ${OBJECT_FILES}
            $OBJCOPY -O ihex radio-rx.elf radio-rx.hex
            '';
          installPhase = ''
            mkdir $out
            cp radio-rx.elf radio-rx.hex $out
            '';
        };

      packages."x86_64-linux"."interrupts-standalone" = (packages."x86_64-linux"."interrupts" atmega32u4);

      packages."x86_64-linux"."interrupts" =
        { mcu, chip_select, f_cpu }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.stdenv.mkDerivation rec {
          name = "interrupts";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."dio" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."shift-register" { inherit mcu chip_select f_cpu; })
            (packages."x86_64-linux"."display" { inherit mcu chip_select f_cpu; })
          ];

          CFLAGS = mcu_cflags { inherit mcu chip_select f_cpu; };
          INCLUDE_DIRS = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);

          buildPhase = ''
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o main.o -c ${src}/interrupts/main.c
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} -o interrupts.elf main.o ${OBJECT_FILES}
            $OBJCOPY -O ihex interrupts.elf interrupts.hex
            '';
          installPhase = ''
            mkdir $out
            cp interrupts.elf interrupts.hex $out
            '';
        };

      packages."x86_64-linux"."power-management-standalone" = (packages."x86_64-linux"."power-management" attiny85);

      packages."x86_64-linux"."power-management" =
        { mcu, chip_select, f_cpu }:
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          lib = pkgs.lib;
          avr = pkgs.pkgsCross.avr.buildPackages;
        in pkgs.stdenv.mkDerivation rec {
          name = "power-management";
          src = ./.;

          buildInputs = [
            (packages."x86_64-linux"."dio" { inherit mcu chip_select f_cpu; })
          ];

          CFLAGS = mcu_cflags { inherit mcu chip_select f_cpu; };
          INCLUDE_DIRS = pkgs.lib.concatStringsSep " " (map (dir: "-I${dir}") buildInputs);
          OBJECT_FILES = pkgs.lib.concatStringsSep " " (map (dir: "${dir}/*.o") buildInputs);

          buildPhase = ''
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} ${INCLUDE_DIRS} -o main.o -c ${src}/power-management/main.c
            ${avr.gcc}/bin/avr-gcc ${CFLAGS} -o power-management.elf main.o ${OBJECT_FILES}
            $OBJCOPY -O ihex power-management.elf power-management.hex
            '';
          installPhase = ''
            mkdir $out
            cp power-management.elf power-management.hex $out
            '';
        };
        */

      devShell."x86_64-linux" =
        let
          pkgs = import nixpkgs { system = "x86_64-linux"; };
          avr = with pkgs.pkgsCross.avr.buildPackages; [
            binutils
            gcc
            gdb
            avrdude
            simavr
            gtkwave
          ];
        in
        pkgs.mkShell {
          name = "avr-shell";
          buildInputs = avr;

          GCC = pkgs.pkgsCross.avr.buildPackages.gcc;
          SIMAVR = pkgs.pkgsCross.avr.buildPackages.simavr;
        };
    };
}
